#!/usr/bin/env python3

# Can be run in three modes: first mode sends commands to the daemon, like 'file transfer', 'notification', etc.
# Obviously a target host should be specified, where the second 'discovery' mode comes in handy
# The third mode is a 'listening' mode where it listens to incoming notifications, progress info, errors, etc.
# Required 2 UNIX sockets to be configured - one for the commands, second for listening to incoming messages
# A custom implementation of Pattern can be used to add some more sophisticated logic for incoming data 

import os
import sys
import argparse
import selectors
import traceback

from pathlib import Path

from protocol import Pattern
from protocol import Missive
from protocol import NoReply
from daemon import Env
from daemon import Defaults
from util import accept_connection
from util import send_via_unix
from util import register_unix_handler

_COMMAND_SOKET = None
_EVENT_SOCKET = None

LINE_UP = '\033[1A'
LINE_CLEAR = '\x1b[2K'


class PrintOutDiscovery(Pattern):
    """Client pattern for printing discovery info."""

    def onprogress(self, handler, missive):
        pass

    def onsent(self, handler, missive):
        pass

    def onreceived(self, handler, missive):
        for k, v in missive.data.items():
            print(f'{k}\t{v}')
        if handler.canclose:
            handler.close()


class PrintOutEvent(Pattern):
    """Client pattern for printing events in console."""

    def onprogress(self, handler, missive):
        pass

    def onsent(self, handler, missive):
        pass

    def onreceived(self, handler, missive):
        if handler.canclose:
            handler.close()
        action = missive.data.get('action', None)
        if not action:
            return
        elif action == 'onprogress':
            total = missive.data.get('total', 1)
            now = missive.data.get('processed', 0)
            fname = missive.data.get('fname', 'unknown')
            way = missive.data.get('way', '').upper()
            print(f'{way} {fname} {now * 100 / total:.2f}%')
            print(LINE_UP, end=LINE_CLEAR)
        elif action == 'onnotify':
            notification = missive.data.get('txt')
            print(notification)
        elif action == 'onerror':
            notification = missive.data.get('txt')
            print(f'ERROR {notification}')
        else:
            pass


def get_run_options() -> dict:
    result = dict(file='', message='', target='')
    parser = argparse.ArgumentParser(
        prog='tnet',
        description='Sends files or messages to another machine')
    parser.add_argument('--file',
                        '-f',
                        type=str,
                        help='Full(!) path to a file to be sent',
                        required=False)
    parser.add_argument('--message',
                        '-m',
                        type=str,
                        help='A message to be sent',
                        required=False)
    parser.add_argument('--target',
                        '-t',
                        type=str,
                        help='Local IP address of a target machine (IPv4)',
                        required=False)
    parser.add_argument('--discover',
                        '-d',
                        action='store_true',
                        help='Checks local network for listening ' +
                        'daemons on other machines',
                        required=False)
    parser.add_argument('--listen',
                        '-l',
                        action='store_true',
                        help='Start listening a UNIX socket for ' +
                        'events from the local daemon',
                        required=False)
    args = parser.parse_args()
    result['listen'] = args.listen
    result['discover'] = args.discover
    result['target'] = args.target
    if not args.file and not args.message:
        result['message'] = '<Hello world!>'
    else:
        result['file'] = args.file
        result['message'] = args.message

    return result


def loop(sel) -> int:
    exit_code = 0
    try:
        while True:
            events = sel.select(timeout=None)
            for key, mask in events:
                handler = key.data
                try:
                    handler.handle_events(mask)
                except Exception as ex:
                    exit_code = 1
                    print(
                        f'Exception for {handler.addr}, {ex}:\n'
                        f'{traceback.format_exc()}',
                        file=sys.stderr)
                    handler.close()
            if not sel.get_map():
                break
    except KeyboardInterrupt:
        exit_code = 1
        print('Caught keyboard interrupt, exiting', file=sys.stderr)
    finally:
        sel.close()
    return exit_code


def recycle_old_sockets():
    """Sockets will be created on bind(), however the directory tree must exist.

    Old files have to be removed to prevent data corruption.
    """
    global _EVENT_SOCKET
    evt_file = Path(_EVENT_SOCKET)
    evt_file.parent.mkdir(parents=True, exist_ok=True)
    evt_file.unlink(missing_ok=True)


def as_events_listener(sel):
    global _EVENT_SOCKET
    _EVENT_SOCKET = os.environ.get(Env.EVENT_SOCKET, Defaults.EVENT_SOCKET) 
    recycle_old_sockets()
    register_unix_handler(sel, _EVENT_SOCKET)
    try:
        while True:
            events = sel.select(timeout=None)
            for key, mask in events:
                if key.data is None:
                    accept_connection(sel, key.fileobj, PrintOutEvent())
                else:
                    try:
                        key.data.handle_events(mask)
                    except Exception:
                        print(
                            f'Handling failed, connection will be closed: \n'
                            f'{traceback.format_exc()}',
                            file=sys.stderr)
                        key.data.close()
    except KeyboardInterrupt:
        print('Caught interrupt, exiting', file=sys.stderr)
    finally:
        print('Closing selector', file=sys.stderr)
        sel.close()


if __name__ == '__main__':
    _COMMAND_SOKET = os.environ.get(Env.COMMAND_SOKET, Defaults.COMMAND_SOKET)

    if not Path(_COMMAND_SOKET).exists():
        sys.exit(f'socket file {_COMMAND_SOKET} not found')
    params = get_run_options()
    payloads = []
    listen_mode = params['listen']
    discovery_mode = params['discover']
    filesend_mode = params['file'] and params['target'] and Path(params['file']).exists()
    notification_mode = params['message'] and params['target']

    sel = selectors.DefaultSelector()
    if listen_mode:
        sys.exit(as_events_listener(sel))
    elif discovery_mode:
        print('appending a discover command')
        data = dict(action='discover')
        payloads.append(Missive(data))
    elif filesend_mode:
        print(f'appending a file {params["file"]}')
        data = dict(action='filetransfer',
                    filepath=params['file'],
                    target=params['target'])
        payloads.append(Missive(data))
    elif notification_mode:
        print(f'appending a msg {params["message"]}')
        data = dict(action='notify',
                    value=params['message'],
                    target=params['target'])
        payloads.append(Missive(data))
    else:
        sys.exit('try to run it with the "--help" option first')

    try:
        if discovery_mode:
            send_via_unix(sel, _COMMAND_SOKET, PrintOutDiscovery(), *payloads)
        else:
            send_via_unix(sel, _COMMAND_SOKET, NoReply(), *payloads)
    except ConnectionRefusedError:
        sys.exit('Connection refused via UNIX socket, check if local daemon is running')
    except Exception as e:
        sys.exit(f'Connection failed via UNIX socket: {e}')
    else:
        sys.exit(loop(sel))
