#!/usr/bin/env python3

# Lisnens for incoming connections using the selectors module, once an connections is established,
# executes some action like accepting a file transfer, relaying a notification to the dedicated UNIX socket,
# replying with the host name and its IPv4 address, etc.
# While sending a file, the progress info can be optionally posted to the dedicated UNIX socket (called events socket)
# for further processing by an external app
# Class called LordCommander plays a role of an orchestrator and contains all the main logic, and also is a Pattern itself
# Due to the use of the selectors module, the daemon is asyncronos and single-threaded

import os
import sys
import socket
import selectors
import traceback
import signal
import shutil

from pathlib import Path

from protocol import Pattern
from protocol import Missive
from protocol import Parcel
from protocol import NoReply

from util import register_tcp_handler
from util import register_unix_handler
from util import register_udp_handler
from util import send_via_tcp
from util import broadcast_via_udp
from util import send_via_unix
from util import accept_connection

_COMMAND_SOCKET = None
_EVENT_SOCKET = None
_BROACAST_PORT = None
_RECEIVE_PORT = None
_FILE_SAVE_DIR = None


class SupportedActions:
    DISCOVER = 'discover'
    BROADCAST = 'broadcast'
    NOTIFY = 'notify'
    QUERY = 'query'
    FILETRANSFER = 'filetransfer'
    EXPOSE = 'expose'
    EXPOSED = 'exposed'


class Env:
    EVENT_SOCKET = 'EVENT_SOCKET'
    COMMAND_SOKET = 'EVENT_SOCKET'
    BROADCAST_PORT = 'BROADCAST_PORT'
    RECEIVE_PORT = 'RECEIVE_PORT'
    FILE_SAVE_DIR = 'FILE_SAVE_DIR'
    REPORT_PROGRESS_IN = 'REPORT_PROGRESS_IN'
    REPORT_PROGRESS_OUT = 'REPORT_PROGRESS_OUT'


class Defaults:
    EVENT_SOCKET = './run/tnet/events'
    COMMAND_SOKET = './run/tnet/commands'
    BROADCAST_PORT = 0xC0DE
    RECEIVE_PORT = 0xC00D
    FILE_SAVE_DIR = 'Downloads'
    REPORT_PROGRESS_IN = 0
    REPORT_PROGRESS_OUT = 0


class LordCommander(Pattern):
    """Server pattern."""

    def __init__(self, *, lan_ip, broadcast_lan_ip, main_selector,
                 broadcast_port, receive_port,
                 file_save_dir, event_socket_path,
                 report_incoming_progress, report_outcoming_progress):
        self.lan_ip = lan_ip
        self.broadcast_lan_ip = broadcast_lan_ip
        self.main_selector = main_selector
        self.broadcast_port = broadcast_port
        self.receive_port = receive_port
        self.file_save_dir = file_save_dir
        self.event_socket_path = event_socket_path
        self.report_incoming_progress = report_incoming_progress
        self.report_outcoming_progress = report_outcoming_progress
        self._actions = {
            SupportedActions.DISCOVER: self._discover,
            SupportedActions.BROADCAST: self._broadcast,
            SupportedActions.NOTIFY: self._notify,
            SupportedActions.QUERY: self._query,
            SupportedActions.FILETRANSFER: self._filetransfer,
            SupportedActions.EXPOSE: self._expose,
            SupportedActions.EXPOSED: self._exposed
        }
        self._discovered = {}
        self._progress = {}

    def onprogress(self, handler, missive):
        if not self.event_socket_path:
            return
        if not self.report_incoming_progress:
            return

        f_name = missive.jsonheader.get('filename', 'unknown')
        total = (missive.jsonheader.get('content-length', 1)
                 if missive.jsonheader else 1)
        processed = (missive.payload_processed_b
                     if missive.payload_processed_b else 0)
        pct = int(processed * 100 / total)
        if missive.missive_id not in self._progress and pct != 100:
            self._progress[missive.missive_id] = pct
        if self._progress[missive.missive_id] == pct:
            return
        else:
            self._progress[missive.missive_id] = pct

        try:
            send_via_unix(self.main_selector,
                          self.event_socket_path,
                          NoReply(),
                          Missive(dict(action='onprogress',
                                       total=total,
                                       processed=processed,
                                       way='in',
                                       fname=f_name,
                                       mid=str(missive.missive_id))))
        except (FileNotFoundError, ConnectionRefusedError):
            pass

        if pct == 100:
            del self._progress[missive.missive_id]

    def onsent(self, handler, missive):
        if handler.canclose:
            handler.close()
        if missive.error:
            self._report_error(handler, missive)

    def onreceived(self, handler, missive):
        if missive.error:
            self._report_error(handler, missive)
        elif missive.jsonheader['content-type'] == 'text/json':
            action = missive.data.get('action')
            if action is None or action not in self._actions:
                self._notify(handler, missive)
            else:
                self._actions.get(action)(handler, missive)
        elif missive.jsonheader['content-type'] == 'file':
            if handler.canclose:
                handler.close()
            orig_name = missive.jsonheader['filename']
            curr_name = missive.data
            Path(self.file_save_dir).mkdir(parents=True, exist_ok=True)
            shutil.move(curr_name,
                        Path(self.file_save_dir) / orig_name)
        else:
            self._notify(handler, missive)

    def _filetransfer(self, handler, missive):
        if handler.canclose:
            handler.close()
        target = missive.data.get('target', '127.0.0.1')
        path = Path(missive.data.get('filepath', ''))
        if not path.is_file():
            print(f'File {path} not found (ensure "filepath" is provided)',
                  file=sys.stderr)
            return
        if target in ['localhost', '127.0.0.1', self.lan_ip]:
            shutil.copy2(path, path.home / self.file_save_dir / path.name)
        else:
            pattern = (ReportSendProgress(self.main_selector, self.event_socket_path)
                       if self.report_outcoming_progress else NoReply())
            send_via_tcp(self.main_selector, target, self.receive_port,
                         pattern,
                         Parcel(path))

    def _discover(self, handler, missive):
        broadcast_via_udp(self.main_selector, self.broadcast_lan_ip,
                          self.broadcast_port, Missive(dict(
                            action=SupportedActions.EXPOSE,
                            replyto=self.lan_ip)))
        # write back whatever we have at the moment (can be nothing)
        handler.enqueue(Missive(self._discovered))
        handler.set_selector_events_mask('w')

    def _expose(self, handler, missive):
        print(missive)
        target = missive.data.get('replyto', None)
        if target == self.lan_ip:
            return
        payload = dict(action=SupportedActions.EXPOSED,
                       ip=self.lan_ip,
                       host=socket.gethostname())
        send_via_tcp(self.main_selector, target, self.receive_port,
                     NoReply(), Missive(payload))

    def _exposed(self, handler, missive):
        if handler.canclose:
            handler.close()
        ip = missive.data.get('ip', None)
        name = missive.data.get('host', None)
        if ip:
            self._discovered[ip] = name

    def _notify(self, handler, missive):
        if handler.canclose:
            handler.close()
        if isinstance(missive.data, dict):
            payload = missive.data.get('value', missive.data)
            target = missive.data.get('target', None)
            if target and payload:
                send_via_tcp(self.main_selector, target, self.receive_port,
                             NoReply(), Missive(payload))
            else:
                self._accept_notification(payload)
        else:
            self._accept_notification(missive.data)

    def _accept_notification(self, notification):
        if not self.event_socket_path:
            return

        try:
            send_via_unix(self.main_selector,
                          self.event_socket_path,
                          NoReply(),
                          Missive(dict(action='onnotify',
                                       txt=notification)))
        except (FileNotFoundError, ConnectionRefusedError):
            pass

    def _report_error(self, handler, missive):
        if handler.canclose:
            handler.close()

        if missive.jsonheader['content-type'] == 'file' and missive.data:
            Path(missive.data).unlink()

        if not self.event_socket_path:
            return
        try:
            send_via_unix(self.main_selector,
                          self.event_socket_path,
                          NoReply(),
                          Missive(dict(action='onerror',
                                       txt=missive.error)))
        except (FileNotFoundError, ConnectionRefusedError):
            pass

    def _broadcast(self, handler, missive):
        pass

    def _query(self, handler, missive):
        payload = missive.data.get('value')
        if not payload:
            handler.close()
        else:
            handler.enqueue(Missive(f'replying to: {payload}'))
            handler.set_selector_events_mask('w')


class ReportSendProgress(Pattern):
    """Pattern that writes progress status to event socket."""

    def __init__(self, selector, event_socket_path):
        self._selector = selector
        self._socket_path = event_socket_path
        self._sent_pct = 0

    def onprogress(self, handler, missive):
        if not self._socket_path:
            return
        if not isinstance(missive, Parcel):
            return

        total = missive.size if missive.size else 1
        processed = (missive.payload_processed_b
                     if missive.payload_processed_b else 0)
        pct = int(processed * 100 / total)
        if pct == self._sent_pct:
            return
        else:
            self._sent_pct = pct

        try:
            send_via_unix(self._selector,
                          self._socket_path,
                          NoReply(),
                          Missive(dict(action='onprogress',
                                       total=total,
                                       processed=processed,
                                       way='out',
                                       fname=missive.filename,
                                       mid=str(missive.missive_id))))
        except (FileNotFoundError, ConnectionRefusedError):
            pass

    def onsent(self, handler, missive):
        if handler.canclose:
            handler.close()

        if not self._socket_path or not isinstance(missive, Parcel):
            return
        notification = {}
        if missive.error:
            notification = dict(action='onerror',
                                txt=f'{missive.path}: {missive.error}')
        else:
            notification = dict(action='onnotify',
                                txt=f'{missive.path} has been sent')
        try:
            send_via_unix(self._selector,
                          self._socket_path,
                          NoReply(),
                          Missive(notification))
        except (FileNotFoundError, ConnectionRefusedError):
            pass

    def onreceived(self, handler, missive):
        pass


def signal_handler(sig, frame):
    raise KeyboardInterrupt


def get_lan_ips():
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.connect(('<broadcast>', 65535))
        lan_ip = sock.getsockname()[0]
    chunks = lan_ip.split('.')
    chunks[3] = '255'
    lan_broadcast_ip = '.'.join(chunks)
    return (lan_ip, lan_broadcast_ip)


def recycle_old_sockets():
    """Sockets will be created on bind(), however the directory tree must exist.

    Old files have to be removed to prevent data corruption.
    """
    global _COMMAND_SOCKET
    cmd_file = Path(_COMMAND_SOCKET)
    cmd_file.parent.mkdir(parents=True, exist_ok=True)
    cmd_file.unlink(missing_ok=True)


if __name__ == '__main__':

    # register handlers for INTERRUPT and TERMINATE signals
    # SIGKILL cannot be handled
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    _COMMAND_SOCKET = os.environ.get(Env.COMMAND_SOKET, Defaults.COMMAND_SOKET)
    _EVENT_SOCKET = os.environ.get(Env.EVENT_SOCKET, Defaults.EVENT_SOCKET)
    _BROACAST_PORT = os.environ.get(Env.BROADCAST_PORT,
                                    Defaults.BROADCAST_PORT)
    _RECEIVE_PORT = os.environ.get(Env.RECEIVE_PORT, Defaults.RECEIVE_PORT)
    _FILE_SAVE_DIR = os.environ.get(Env.FILE_SAVE_DIR, Defaults.FILE_SAVE_DIR)
    _REPORT_PROGRESS_IN = os.environ.get(Env.REPORT_PROGRESS_IN,
                                         Defaults.REPORT_PROGRESS_IN)
    _REPORT_PROGRESS_OUT = os.environ.get(Env.REPORT_PROGRESS_OUT,
                                          Defaults.REPORT_PROGRESS_OUT)

    recycle_old_sockets()

    lan_ip, broadcast_lan_ip = get_lan_ips()
    sel = selectors.DefaultSelector()

    # a bunch of listeners to be used for communication
    udp_handler = register_udp_handler(sel, broadcast_lan_ip, _BROACAST_PORT)
    register_unix_handler(sel, _COMMAND_SOCKET)
    register_tcp_handler(sel, lan_ip, _RECEIVE_PORT)

    orchestrator = LordCommander(
        lan_ip=lan_ip,
        broadcast_lan_ip=broadcast_lan_ip,
        main_selector=sel,
        broadcast_port=_BROACAST_PORT,
        receive_port=_RECEIVE_PORT,
        file_save_dir=_FILE_SAVE_DIR,
        event_socket_path=_EVENT_SOCKET,
        report_incoming_progress=_REPORT_PROGRESS_IN,
        report_outcoming_progress=_REPORT_PROGRESS_OUT
    )

    udp_handler.setpattern(orchestrator)

    try:
        while True:
            events = sel.select(timeout=None)
            for key, mask in events:
                if key.data is None:
                    accept_connection(sel, key.fileobj, orchestrator)
                else:
                    try:
                        key.data.handle_events(mask)
                    except Exception:
                        print(
                            f'Handling failed, connection will be closed: \n'
                            f'{traceback.format_exc()}',
                            file=sys.stderr)
                        key.data.close()
    except KeyboardInterrupt:
        print('Caught interrupt, exiting', file=sys.stderr)
    finally:
        sel.close()
