# Missive and Parcel classes are 'data transfer' entities which contol
# encoding/decoding, length of an transfered object,
# whereas DefaultHandler contains logic for sending/receiving them
# using selectors - thus it controls when a socket can be closed,
# when to switch the selector from write to read mode, etc.
# There is also a concept of a 'Pattern', a concrete implementation of which
# executes some logic on 3 events: onsent, onreceived, and onprogress.
# for example NoReply pattern closes the handler in onsent() which means
# once a message or a file was sent, no reply is expected and we are done here


import sys
import selectors
import json
import io
import struct
import socket
from abc import ABC, abstractmethod
import tempfile
import os
import uuid


class Missive:

    def __init__(self, data=None):
        self.data = data
        self.jsonheader = None
        self.payload_processed_b = 0
        self._complete = False
        self._recv_buffer = b''
        self._send_buffer = self._encode()
        self._jsonheader_len = None
        self.missive_id = uuid.uuid4()
        self.error = ''
        # used for files only
        self._file_total_recv = 0
        self._file = None

    def __str__(self):
        return f'{self.__class__.__name__} <{self.jsonheader} {self.data or "None"}>'

    def reset(self):
        if self._recv_buffer:
            raise ProtocolError('Resseting of incoming is not possible')
        self.jsonheader = None
        self.payload_processed_b = 0
        self.missive_id = uuid.uuid4()
        self.error = ''
        self._complete = False
        self._recv_buffer = b''
        self._send_buffer = self._encode()
        self._jsonheader_len = None

    @property
    def complete(self):
        return self._complete

    def consume_bytes(self, bytes):
        # print(f'got {len(bytes)} bytes')
        if self.complete:
            return bytes
        consumed = 0
        self._recv_buffer += bytes
        # print(f'recv_buffer {len(self._recv_buffer)} bytes')
        if self._jsonheader_len is None:
            consumed += self._process_protoheader()

        if self._jsonheader_len is not None:
            if self.jsonheader is None:
                consumed += self._process_jsonheader()

        if self.jsonheader:
            consumed += self._process_payload(bytes)

        bytes = bytes[consumed:]
        return bytes

    def _process_protoheader(self):
        hdrlen = 2
        read = 0
        if len(self._recv_buffer) >= hdrlen:
            self._jsonheader_len = struct.unpack('>H',
                                                 self._recv_buffer[:hdrlen])[0]
            self._recv_buffer = self._recv_buffer[hdrlen:]
            read = hdrlen
        return read

    def _process_jsonheader(self):
        hdrlen = self._jsonheader_len
        read = 0
        if len(self._recv_buffer) >= hdrlen:
            self.jsonheader = self._json_decode(self._recv_buffer[:hdrlen],
                                                'utf-8')
            self._recv_buffer = self._recv_buffer[hdrlen:]
            for reqhdr in (
                    'byteorder',
                    'content-length',
                    'content-type',
                    'content-encoding',
            ):
                if reqhdr not in self.jsonheader:
                    raise ProtocolError(f'Missing required header "{reqhdr}"')
            read = hdrlen
        return read

    def _process_payload(self, bytes):
        if self.jsonheader['content-type'] == 'file':
            return self._process_file(bytes)
        else:
            return self._process_text(bytes)

    def _process_text(self, bytes):
        read = 0
        content_len = self.jsonheader['content-length']
        if not len(self._recv_buffer) >= content_len:
            read = len(bytes)
        else:
            content = self._recv_buffer[:content_len]
            read = content_len
            self._recv_buffer = b''
            if self.jsonheader['content-type'] == 'text/json':
                encoding = self.jsonheader['content-encoding']
                self.data = self._json_decode(content, encoding)
            elif self.jsonheader['content-type'] == 'text/plain':
                encoding = self.jsonheader['content-encoding']
                self.data = content.decode(encoding)
            else:
                # Binary or unknown content-type
                self.data = content
            self._complete = True
        self.payload_processed_b += read
        return read

    def _process_file(self, bytes):
        read = 0
        content_len = self.jsonheader['content-length']
        if self._file_total_recv < content_len:
            if not self._file:
                self.data = os.path.join(tempfile.gettempdir(),
                                         f'{uuid.uuid4()}')
                self._file = open(self.data, 'wb')
            read = self._file.write(self._recv_buffer)
            self._recv_buffer = self._recv_buffer[read:]
        else:
            self._file.close()
            self._file = None
            self._complete = True
        self._file_total_recv += read
        self.payload_processed_b += read
        return read

    def yield_bytes(self, consumer):
        consumed = self._send_from_buffer(consumer)
        if len(self._send_buffer) == 0:
            self._complete = True
        return consumed

    def _send_from_buffer(self, consumer):
        if self._complete:
            return 0
        if len(self._send_buffer) == 0:
            return 0
        chunk = b''
        if len(self._send_buffer) > 4096:
            chunk = self._send_buffer[:4096]
        else:
            chunk = self._send_buffer
        consumed = consumer(chunk)
        if consumed is None:
            raise ProtocolError('Bytes consumer is of an inappropriate type')
        self._send_buffer = self._send_buffer[consumed:]
        return consumed

    def _encode(self):
        if self.data is None:
            return b''
        payload = None
        jsonheader = {
            'byteorder': sys.byteorder,
            'content-type': None,
            'content-encoding': 'utf-8',
            'content-length': 0,
        }
        if type(self.data) is str:
            jsonheader['content-type'] = 'text/plain'
            payload = self.data.encode('utf-8')
        elif type(self.data) is dict:
            jsonheader['content-type'] = 'text/json'
            payload = self._json_encode(self.data, 'utf-8')
        else:
            raise ProtocolError('Supported data types: str, dict')

        jsonheader['content-length'] = len(payload)
        jsonheader_bytes = self._json_encode(jsonheader, 'utf-8')
        missive_hdr = struct.pack('>H', len(jsonheader_bytes))
        encoded = missive_hdr + jsonheader_bytes + payload
        return encoded

    def _json_encode(self, obj, encoding):
        return json.dumps(obj, ensure_ascii=False).encode(encoding)

    def _json_decode(self, json_bytes, encoding):
        tiow = io.TextIOWrapper(io.BytesIO(json_bytes),
                                encoding=encoding,
                                newline='')
        obj = json.load(tiow)
        tiow.close()
        return obj


class Parcel(Missive):
    """Use it for sending a file."""

    def __init__(self, path):
        self.path = path
        self.filename = ''
        self.size = 0
        super().__init__()

    def yield_bytes(self, consumer):
        consumed = self._send_from_buffer(consumer)
        consumed += self._send_file(consumer)
        return consumed

    def _send_file(self, consumer):
        if self._complete:
            return 0
        if not self._file:
            self._file = open(self.path, 'rb')

        chunk = self._file.read(4096)
        if chunk == b'':
            self._file.close()
            self._file = None
            self._complete = True
            consumed = 0
        else:
            consumed = consumer(chunk)
            if consumed is None:
                raise ProtocolError(
                    'Bytes consumer is of an inappropriate type')
            if consumed < len(chunk):
                diff = len(chunk) - consumed
                self._file.seek(-diff, 1)
        self.payload_processed_b += consumed
        return consumed

    def _encode(self):
        if self.path and os.path.isfile(self.path):
            self.filename = os.path.basename(self.path)
            self.size = os.path.getsize(self.path)
        jsonheader = {
            'byteorder': sys.byteorder,
            'content-type': 'file',
            'content-encoding': 'raw',
            'filename': self.filename,
            'content-length': self.size,
        }

        jsonheader_bytes = self._json_encode(jsonheader, 'utf-8')
        parcel_hdr = struct.pack('>H', len(jsonheader_bytes))
        encoded = parcel_hdr + jsonheader_bytes
        return encoded


class ProtocolError(Exception):

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f'ProtocolError: {self.message}'


class DefaultHandler:

    def __init__(self, selector, sock, addr, tag='handler'):
        self.selector = selector
        self.sock = sock
        self.addr = addr
        self.tag = tag
        self._queue = []
        self._recv = None
        self._callback = NonePattern()
        self._canclose = True
        self._isclosed = False

    def setpattern(self, pattern):
        self._callback = pattern

    def set_selector_events_mask(self, mode):
        """Set selector to listen for events: mode is 'r', 'w', or 'rw'."""

        if self.sock is None:
            return
        if mode == 'r':
            events = selectors.EVENT_READ
        elif mode == 'w':
            events = selectors.EVENT_WRITE
        elif mode == 'rw':
            events = selectors.EVENT_READ | selectors.EVENT_WRITE
        else:
            raise ProtocolError(f'mask mode {mode!r} not recognized')
        self.selector.modify(self.sock, events, data=self)

    def enqueue(self, missive):
        self._queue.append(missive)
        self.set_selector_events_mask('rw')

    def _sendout(self, bytes):
        if self.sock.proto == socket.IPPROTO_UDP:
            return self.sock.sendto(bytes, self.sock.getsockname())
        else:
            return self.sock.send(bytes)

    def _read(self):
        try:
            bytes = self.sock.recv(4096)
        except BlockingIOError as e:
            # Resource temporarily unavailable (errno EWOULDBLOCK)
            # with the next attempt it might be ok - so pass for now
            print(f'BlockingIOError {e}')
        else:
            read = True
            self._canclose = len(bytes) == 0
            while read:
                if not self._recv:
                    self._recv = Missive()
                try:
                    bytes = self._recv.consume_bytes(bytes)
                except Exception as e:
                    self._canclose = True
                    self._recv.error = f'Failed to receive, {e}'
                    self._callback.onreceived(self, self._recv)
                    self._recv = None
                    read = False
                else:
                    if self._recv.complete:
                        self._canclose = len(bytes) == 0
                        self._callback.onreceived(self, self._recv)
                        self._recv = None
                    else:
                        self._callback.onprogress(self, self._recv)     
                    if bytes:
                        self._recv = Missive()
                    read = len(bytes) > 0

    def _write(self):
        if not self._queue or self._queue[0] is None:
            self._queue = []
            self.set_selector_events_mask('r')
            return
        try:
            self._queue[0].yield_bytes(self._sendout)
        except BlockingIOError as e:
            # Resource temporarily unavailable (errno EWOULDBLOCK)
            # with the next attempt it might be ok - so pass for now
            print(f'BlockingIOError {e}', file=sys.stderr)
        except (BrokenPipeError, ConnectionResetError) as e:
            miss = self._queue.pop(0)
            miss.error = f'Counterparty unable to receive data, {e}'
            self._callback.onsent(self, miss)
            self.close()
            print(miss.error, file=sys.stderr)
        else:
            if self._queue[0].complete:
                self._callback.onsent(self, self._queue.pop(0))
            else:
                self._callback.onprogress(self, self._queue[0])

        if not self._queue and not self._isclosed:
            # Set selector to listen for read events, we're done writing.
            self.set_selector_events_mask('r')

    def handle_events(self, mask):
        if mask & selectors.EVENT_WRITE:
            self._write()
        if mask & selectors.EVENT_READ:
            self._read()

    @property
    def canclose(self):
        return len(self._queue) == 0 and self._canclose

    def close(self):
        try:
            if not self._isclosed:
                self.selector.unregister(self.sock)
        except Exception as e:
            print(
                f'Error while unregister the socket: selector.unregister() exception for '
                f'{self.addr}: {e!r}', file=sys.stderr)
        try:
            if not self._isclosed:
                self.sock.close()
        except OSError as e:
            print(
                f'{self.__class__.__name__}. Error: socket.close() exception for {self.addr}: {e!r}',
                file=sys.stderr
            )
        finally:
            # Delete reference to socket object for garbage collection
            self.sock = None
            self._callback = None
            self._isclosed = True
            self._canclose = True
            self._queue = []


class Pattern(ABC):
    """Implement a custom logic based on this class."""

    @abstractmethod
    def onprogress(self, handler, missive):
        pass

    @abstractmethod
    def onsent(self, handler, missive):
        pass

    @abstractmethod
    def onreceived(self, handler, missive):
        pass


class NonePattern(Pattern):

    def onprogress(self, handler, missive):
        pass

    def onsent(self, handler, missive):
        pass

    def onreceived(self, handler, missive):
        pass


class NoReply(Pattern):
    """Simple client pattern that does NOT expect a reply from server."""

    def onprogress(self, handler, missive):
        pass

    def onsent(self, handler, missive):
        if handler.canclose:
            handler.close()

    def onreceived(self, handler, missive):
        pass
